#include <iostream>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "Settings.h"
#include "GaborFilter.h"

using namespace std;
using namespace cv;

pair<Mat, Mat> getGaborKernelReIm(Size kernelSize, double sigma, double theta, double lambda, double gamma, double psi) {
	double sigma_x = sigma;
	double sigma_y = sigma / gamma;
	int nstds = 3;
	int xmin, xmax, ymin, ymax;
	double cosTheta = cos(theta), sinTheta = sin(theta);

	if (kernelSize.width > 0)
		xmax = kernelSize.width / 2;
	else
		xmax = cvRound(max(fabs(nstds*sigma_x*cosTheta), fabs(nstds*sigma_y*sinTheta)));

	if (kernelSize.height > 0)
		ymax = kernelSize.height / 2;
	else
		ymax = cvRound(max(fabs(nstds*sigma_x*sinTheta), fabs(nstds*sigma_y*cosTheta)));
	
	xmin = -xmax;
	ymin = -ymax;

	Mat kernelRe(ymax - ymin + 1, xmax - xmin + 1, CV_32F);
	Mat kernelIm(ymax - ymin + 1, xmax - xmin + 1, CV_32F);

	double scale = 1;
	double ex = -0.5 / (sigma_x * sigma_x);
	double ey = -0.5 / (sigma_y * sigma_y);
	double cscale = CV_PI * 2 / lambda;

	for (int y = ymin; y <= ymax; y++) {
		for (int x = xmin; x <= xmax; x++) {
			double xr = x*cosTheta + y*sinTheta;
			double yr = -x*sinTheta + y*cosTheta;
			double vRe = scale * exp(ex * xr * xr + ey * yr * yr) * cos(cscale * xr + psi);
			double vIm = scale * exp(ex * xr * xr + ey * yr * yr) * sin(cscale * xr + psi);
			kernelRe.at<float>(ymax - y, xmax - x) = (float)vRe;
			kernelIm.at<float>(ymax - y, xmax - x) = (float)vIm;
		}
	}
	return make_pair(kernelRe, kernelIm);
}

void fillGaborBank(vector<pair<Mat, Mat>>& gaborBank) {
	double sigma;				// rozptyl
	double theta;				// natoceni
	double lambda;				// vlnova delka
	double gamma = gGamma;      // elipticnost
	double psi = gPsi;          // fazovy posun

	double sigmaLambda;

	for (double bandwidth = gBandwidthMin; bandwidth <= gBandwidthMax; bandwidth += (gBandwidthMax - gBandwidthMin) / ((double)(gBandwidths - 1))) {
		sigmaLambda = (1.0 / CV_PI) * sqrt(log(2.0) / 2.0) * ((pow(2, bandwidth) + 1) / (pow(2, bandwidth) - 1));

		for (lambda = gLabmdaMin; lambda <= gLambdaMax; lambda += (gLambdaMax - gLabmdaMin) / ((double)(gLambdas - 1))) {
			sigma = lambda * sigmaLambda;

			int n = (int)ceil(2.5 * sigma / gamma);
			Size kernelSize(2 * n + 1, 2 * n + 1);

			for (theta = gThetaMin; theta <= gThetaMax; theta += (gThetaMax - gThetaMin) / ((double)(gThetas - 1))) {
				pair<Mat, Mat> kernel = getGaborKernelReIm(kernelSize, sigma, theta, lambda, gamma, psi);
				gaborBank.push_back(kernel);
			}
		}
	}
}

void gaborFilter(Mat& in, Mat& out, vector<pair<Mat, Mat>>& gaborBank) {
	int i, j, k;

	Size gaborSize(gWidth, gHeight);
	Size bankSize(gWidth, (gHeight * gaborBank.size()));

	Mat image;
	in.convertTo(image, CV_32F);
	Mat filtered(bankSize, CV_32F);

	for (i = 0; i < gaborBank.size(); i++) {
		pair<Mat, Mat> kernel = gaborBank.at(i);
		Mat kernelRe = kernel.first;
		Mat kernelIm = kernel.second;
		Mat reF(gaborSize, CV_32F);
		Mat imF(gaborSize, CV_32F);
		Mat magnitude(gaborSize, CV_32F);
		filter2D(image, reF, CV_32F, kernelRe);
		filter2D(image, imF, CV_32F, kernelIm);
		pow(reF, 2, reF);
		pow(imF, 2, imF);
		add(imF, reF, magnitude);
		sqrt(magnitude, magnitude);

		for (j = 0; j < gaborSize.height; j++) {
			for (k = 0; k < gaborSize.width; k++) {
				filtered.at<float>(j + (i * gaborSize.height), k) = magnitude.at<float>(j, k);
			}
		}
	}
	filtered.copyTo(out);
}




    


  
